#-------------------------------------------------------------------------------
# Copyright 2018 Visual Components Ltd. All rights reserved.                   |
#-------------------------------------------------------------------------------
#                    |
#  Version 1.00      |
#                    |
#---------------------
#
#Release notes:
#
#-Supports Yaskawa robots in eCat that use DX100 or DX200 controller
#
#-Supports pulse and rectangular jobs.
#  -By default pp produces rectangular jobs. Pulse jobs can be made if robot component has:
#     -Boolean Property 'UsePulses' set to True.
#     -String property 'PulsesPerRev' which contains gear ratios (pulses per axis revolution) as a list separated with ';' e.g. '100000;200000;300000;400000;500000;600000'
#       -PPR values can be calculated from robot's RC1G parameters (in ALL.prm) 972-977 (gear reduction ratio denominators) and 980-985 (gear reduction ratio numerators). 
#        Value is 4096*num/denom (4096 encoder pulses per motor revolution). Note that sign can be negative if model's positive axis direction is different than on real robot.
#          -S=4096*RC1G980/RC1G972
#          -L=4096*RC1G981/RC1G973
#          -U=4096*RC1G982/RC1G974
#          -R=4096*RC1G983/RC1G975
#          -B=4096*RC1G984/RC1G976
#          -T=4096*RC1G985/RC1G977
#     -String property 'PulsesZeros' which contains joint zero offsets (in pulses) as a list separated with ';' e.g. '0;0;0;0;250000;0'
#       -PPZ values are zeroes if robot model's zero positions are set correctly.
#  -In rectangular jobs all motions in a routine should use the same base frame. If not so, first base frame is used for the job and all motions are converted to that frame.
#     -*Null* and UFRAME0 bases are considered as ROBOT Frame in job file.
#     -To force jobs to be converted into a some frame create string property 'ConvertToFrame' to robot and set value to 'ROBOT' or one of the base frame names (e.g. UFRAME1).
#
#-Supports Position Level to some degree.
#   -To activate PL create boolean property 'UsePL' to robot and set it True.
#     -For motions with zero acceleration PL is not used
#     -PL is either 0 for Accuracy value 0 and 1 otherwise
#
#-Creates one job file per routine. Main routine's file name is selected with dialog, other files are created with names of the subroutines.
#
#-Creates a support files TOOL.CND and UFRAME.CND which contains TOOL and UFRAME data for used frames.
#
#-Supports external axis to some degree.
#  -Track and positioner are supported. Servo tools are not supported.
#  -For pulse jobs external axis components should have these these properties:
#     -String property 'PulsesPerRev' with StepValues list which contains gear ratios (pulses per axis revolution (positioner) / pulses per mm (track)) as a list separated with ';' e.g. '100000;200000'
#     -String property 'PulsesZeros' with StepValues list which contains joint zero offsets (in pulses) as a liset separated with ';' e.g. '0;0'
#  -Supports sync moves.
#     -If base frame is attached to positioner's flange, sync moves are used for linear motions (SMOVL).
#     -Sync moves can be disabled by creating a boolean property 'UseSyncMoves' for robot and setting it False.
#
#End release notes.
#
#---------------------

from vcCommand import *
from math import *
import vcMatrix
import vcVector
import re, os, os.path, sys, string, time
import locale

EPSILON = 1e-6
RAD_TO_DEG = 180.0 / pi
DEG_TO_RAD = pi / 180.0

#-------------------------------------------------------------------------------  
def postProcess(app,program,uri):
  #Post-Processor entry point
  global comp, controller, motiontarget
  global jobfolder
  global usePulses, useSyncMoves, usePL, convertToFrame
  global bases, tools
  global usedBases, usedTools
  global compSTATION, posCmpSTATION, trackCmpSTATION
  global robotAxes, trackAxes, posAxes
  global posCmp, trackCmp
  global ppz, ppr
  
  #This is to keep all numeric conversions consistent across platform locales.
  locale.setlocale(locale.LC_NUMERIC,'C')
  
  #get basic handles
  comp = program.Executor.Component
  controller = program.Executor.Controller
  motiontarget = controller.createTarget()
  currentTarget = controller.createTarget()

  #get job folder
  if uri[len(uri)-8:] == '.JBI.JBI':
    uri = uri[:len(uri)-4]
  head, tail = os.path.split(uri)
  mainName = tail[:len(tail)-4]
  idx = mainName.find( '-' ) 
  if idx > -1:
    mainName = mainName[:idx]
  jobfolder = head + '\\'

  #--
  #Settings
  #--
  #PULSE vs. RECTAN job
  usePulses = False       #Default RECTAN
  prop = comp.getProperty('UsePulses')
  if prop:
    usePulses = prop.Value
  #Use sync moves (SMOVL)
  useSyncMoves = True     #Default on
  prop = comp.getProperty('UseSyncMoves')
  if prop:
    useSyncMoves = prop.Value
  #Position level
  usePL = False           #Default off
  prop = comp.getProperty('UsePL')
  if prop:
    usePL = prop.Value
  #Convert positions to frame in RECTAN job
  convertToFrame = ''
  prop = comp.getProperty('ConvertToFrame')
  if prop:
    convertToFrame = prop.Value

  #get robot joints
  robotAxes = len([ j for j in controller.Joints if not j.ExternalController ])
  trackAxes = 0
  posAxes = 0
  posCmp = None
  trackCmp = None
  ppr = []
  ppz = []
  propPPR = comp.getProperty('PulsesPerRev')
  propPZ  = comp.getProperty('PulsesZeros')
  if propPPR and propPZ:
    sppr = propPPR.Value.split(';')
    sppz = propPZ.Value.split(';')
    if len(sppr) == robotAxes and len(sppz) == robotAxes:
      for j in range(robotAxes):
        try:
          ppr.append( float(sppr[j]) )
          ppz.append( float(sppz[j]) )
        except:
          pass

  #get robot positioner joints
  for bhr in comp.getBehavioursByType(VC_ONETOONEINTERFACE):
    if bhr.IsAbstract:
      continue
    trackCmp = bhr.ConnectedComponent
    trackAxes = 0
    if trackCmp:
      ctrls = trackCmp.findBehavioursByType(VC_SERVOCONTROLLER)
      if ctrls:
        trackCnt = ctrls[0]
        if trackCnt:
          trackAxes = trackCnt.JointCount
          if trackAxes:
            propPPR = trackCmp.getProperty('PulsesPerRev')
            propPZ  = trackCmp.getProperty('PulsesZeros')
            if propPPR and propPZ:
              sppr = propPPR.Value.split(';')
              sppz = propPZ.Value.split(';')
              if len(sppr) == trackAxes and len(sppz) == trackAxes:
                for j in range(trackAxes):
                  try:
                    ppr.append( float(sppr[j]) )
                    ppz.append( float(sppz[j]) )
                  except:
                    pass

  #get workpiece positioners joints
  for bhr in comp.getBehavioursByType(VC_ONETOMANYINTERFACE):
    if not bhr.IsAbstract:
      continue
    if not bhr.ConnectedComponent:
      continue
    posCmp = bhr.ConnectedComponent
    posAxes = 0
    if posCmp:
      posCnt = posCmp.findBehavioursByType(VC_SERVOCONTROLLER )[0]
      if posCnt:
        posAxes += posCnt.JointCount
        if posAxes:
          propPPR = posCmp.getProperty('PulsesPerRev')
          propPZ  = posCmp.getProperty('PulsesZeros')
          if propPPR and propPZ:
            sppr = propPPR.Value.split(';')
            sppz = propPZ.Value.split(';')
            if len(sppr) == posAxes and len(sppz) == posAxes:
              for j in range(posAxes):
                try:
                  ppr.append( float(sppr[j]) )
                  ppz.append( float(sppz[j]) )
                except:
                  pass
  
  #
  if len(ppr) < len(motiontarget.JointValues) or len(ppz) < len(motiontarget.JointValues):
    if usePulses:
      print 'Using pulses but PPR/PPZ data is invalid! Using default PPR/PPZ values.'
      print 'ppr:', len(ppr)
      print 'ppz:', len(ppz)
      print 'motiontarget.JointValues:', len(motiontarget.JointValues)
    ppr = []
    ppz = []
    for i in range(len(motiontarget.JointValues)):
      ppr.append(100000)
      ppz.append(0)
      
  #make a list of tools and bases used in this program
  bases = []
  for b in controller.Bases:
    bases.append(b.Name)
  tools = []
  for t in controller.Tools:
    tools.append(t.Name)
  usedBases = []
  usedTools = []

  #get station names for job header
  prop = comp.getProperty( 'STATION' )
  if prop:
    compSTATION = prop.Value
  else:
    compSTATION = 'RB1'
  #track
  if trackCmp:
    prop = trackCmp.getProperty( 'STATION' )
    if prop:
      trackCmpSTATION = prop.Value
    else:
      trackCmpSTATION = 'BS1'
  #positioner
  if posCmp:
    prop = posCmp.getProperty( 'STATION' )
    if prop:
      posCmpSTATION = prop.Value
    else:
      posCmpSTATION = 'ST1'
      
  #main routine
  filenamelist=[]
  filenamelist.append(uri)
  if writeJob(program.MainRoutine,mainName) == False:
    print 'Failed to write mainroutine'
    return False,filenamelist

  #subroutines
  for routine in program.Routines:
    filename = head + '\\' + routine.Name + '.JBI'
    filenamelist.append(filename)
    if writeJob(routine,routine.Name) == False:
      print 'Failed to write subroutine \'%s\'' % routine.Name
      return False,filenamelist

  #return to initial position
  controller.moveImmediate( currentTarget )
  
  #Write frame info
  writeFrameInfo(filenamelist)

  return True,filenamelist
#-------------------------------------------------------------------------------
def writeJob(routine, name):
  #Write a job into a file
  global jobfolder
  global usePulses, useSyncMoves, usePL, convertToFrame
  global compSTATION, trackCmpSTATION, posCmpSTATION
  global robotAxes, trackAxes, posAxes
  global Ccount, BCcount, ECcount
  global cxxxxx, bcxxxxx, ecxxxxx, inst
  global posIndex
  global tsync
  global initbase, inittool, prevtool, prevrconf, prevt4, prevt6
  global pointsWereConverted
  
  jbiname = getJobName(name)
  jbifilename = jobfolder + jbiname + '.JBI'

  try:
    jbi = open(jbifilename,'w')
  except:
    print 'Cannot open file \'%s\' for writing' % jbifilename
    return False

  cxxxxx = []
  bcxxxxx = []
  ecxxxxx = []
  inst = []

  Ccount = 0
  BCcount = 0
  ECcount = 0
  
  posIndex = -1
  tsync = 0
  
  initbase = -2
  inittool = -2
  prevtool = -2
  prevrconf = -1
  prevt4 = -1
  prevt6 = -1
  frame = 'ROBOT'
  
  pointsWereConverted = False

  if convertToFrame:
    initbase = getFrameIndex(convertToFrame, 'UFRAME', bases )
  
  writeStatement = getAllStatementWriters()
  for statement in routine.Statements:
    writeStatement.get(statement.Type,unhandled)(statement)
  
  if initbase < -1:
    initbase = -1
  
  if inittool < -1:
    inittool = -1
  
  #Header
  jbi.write('/JOB\n')
  jbi.write('//NAME %s\n' % jbiname)
  jbi.write('//POS\n')
  jbi.write('///NPOS %i,%i,%i,0,0,0\n' % (Ccount,BCcount,ECcount))
  jbi.write('///TOOL %i\n' % inittool)
  if usePulses:
    jbi.write('///POSTYPE PULSE\n')
    jbi.write('///PULSE\n')
  else:
    jbi.write('///RECTAN\n')
    if initbase == -1 or initbase == 0:
      frame = 'ROBOT'
      jbi.write('///POSTYPE ROBOT\n' )
      if pointsWereConverted:
        print 'Multiple base frames used in job ' + name + '. Coverting points to ROBOT frame.'
    else:
      frame = '%i' % initbase
      jbi.write('///USER %i\n' % initbase)
      jbi.write('///POSTYPE USER\n' )
      if pointsWereConverted:
        print 'Multiple base frames used in job ' + name + '. Coverting points to UFRAME ' + str(initbase) + '.'
  
  for line in cxxxxx:
    jbi.write(line)
  
  if trackAxes:
    if prevtool != 0:
      jbi.write('///TOOL 0\n')
    if not usePulses:
      jbi.write('///RCONF 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0\n' )
    for line in bcxxxxx:
      jbi.write(line)
  
  if posAxes:
    for line in ecxxxxx:
      jbi.write(line)
  
  #Instructions
  jbi.write('//INST\n')
  jbi.write('///DATE %s\n' % (time.strftime('%Y/%m/%d %H:%M')))
  jbi.write('///ATTR SC,RW')
  if not usePulses:
    jbi.write(',RJ')
  jbi.write('\n')
  if not usePulses:
    jbi.write('///FRAME %s\n' % frame )
  jbi.write('///GROUP1 %s' % compSTATION )  
  if trackAxes:
    jbi.write(',%s' % trackCmpSTATION )
  if not useSyncMoves:
    if posCmp:
      jbi.write(',%s' % posCmpSTATION )
    jbi.write('\n')
  else:
    jbi.write('\n')
    if posCmp:
      jbi.write('///GROUP2 %s\n' % posCmpSTATION )
  jbi.write('NOP\n')

  for line in inst:
    jbi.write( line )

  jbi.write('END\n')
  jbi.close()

  return True
#-------------------------------------------------------------------------------
def writeFrameInfo(filenamelist):
  #Creates a support files TOOL.CND and UFRAME.CND which contains TOOL and UFRAME data for used frames.
  global comp, controller
  global usedBases, usedTools

  #UFRAME.CND
  filename = jobfolder + 'UFRAME.CND'
  
  filenamelist.append(filename)

  try:
    file = open(filename,'w')
  except:
    print 'Cannot open file \'%s\' for writing' % filename
    return False
    
  rm = comp.WorldPositionMatrix
  kin = comp.findBehaviour('Kinematics')
  if kin:
    offset_p = kin.getProperty('L01Z')
    if offset_p:
      rm.translateRel(0,0,offset_p.Value)
  irm = vcMatrix.new(rm)
  irm.invert()
  
  for b in controller.Bases:
    if b.Name in usedBases:
      if b.Node:
        bm = b.Node.WorldPositionMatrix * b.PositionMatrix
        m = irm * bm
      else:
        #Reference is robot world 
        #Note, might produce faulty coordinates if robot is on track or world frame has been moved
        m = b.PositionMatrix
      #
      index = getFrameIndex(b.Name,'UFRAME',tools)
      if index < 1:
        #*NULL* or UFRAME0 (=ROBOT frame), skip
        continue
      name = ''
      if len(b.Name) < 6 or b.Name[0:6] != 'UFRAME':
        name = ' ' + goodName(b.Name)
      file.write('//UFRAME %i\n' % index)
      file.write('///NAME%s\n' % name)
      file.write('///TOOL 0\n')
      file.write('///GROUP 1,0,0,0\n')
      file.write('///PULSE\n')
      file.write('////RORG C000=0,0,0,0,0,0\n')
      file.write('////RXX C001=0,0,0,0,0,0\n')
      file.write('////RXY C002=0,0,0,0,0,0\n')
      file.write('////BUSER %.3f,%.3f,%.3f,%.2f,%.2f,%.2f\n' % (m.P.X,m.P.Y,m.P.Z,m.WPR.X,m.WPR.Y,m.WPR.Z))
      
  file.close()
  
  #TOOL.CND
  filename = jobfolder + 'TOOL.CND'
  
  filenamelist.append(filename)
  
  try:
    file = open(filename,'w')
  except:
    print 'Cannot open file \'%s\' for writing' % filename
    return False
    
  t0m = vcMatrix.new()
  try:
    t0m = controller.FlangeNode.WorldPositionMatrix
  except:
    pass
  it0m = vcMatrix.new(t0m)
  it0m.invert()
  
  for t in controller.Tools:
    if t.Name in usedTools:
      if t.Node:
        tm = t.Node.WorldPositionMatrix * t.PositionMatrix
      else:
        tm = t.PositionMatrix
      m = it0m * tm
      #
      index = getFrameIndex(t.Name,'TOOL',tools)
      name = ''
      if len(t.Name) < 4 or t.Name[0:4] != 'TOOL':
        name = ' ' + goodName(t.Name)
      file.write('//TOOL %i\n' % index)
      file.write('///NAME%s\n' % name)
      file.write('%.3f,%.3f,%.3f,%.2f,%.2f,%.2f\n' % (m.P.X,m.P.Y,m.P.Z,m.WPR.X,m.WPR.Y,m.WPR.Z))
      file.write('0.000,0.000,0.001\n')
      file.write('0.001\n')
      file.write('0.000,0.000,0.000\n')
      file.write('0,0\n')
      
  file.close()
#-------------------------------------------------------------------------------
def writeCall(statement, ifclause=''):
  global inst
  inst.append('CALL JOB:%s%s\n' % (goodName(statement.Routine.Name), ifclause) )
#-------------------------------------------------------------------------------
def writeComment(statement, ifclause=''):
  global inst
  inst.append('\'%s\n' % statement.Comment)
#-------------------------------------------------------------------------------
def writeDefineBase(statement, ifclause=''):
  global inst
  relative = ''
  if statement.Relative:
    relative = 'RELATIVE'
  p = statement.Position.P
  e = calcRxRyRz( statement.Position )
  name = statement.Base.Name
  inst.append('\'DEFINE %s BASE %s %g,%g,%g,%g,%g,%g\n' %(relative,name,p.X,p.Y,p.Z,e.X,e.Y,e.Z))
#-------------------------------------------------------------------------------
def writeDefineTool(statement, ifclause=''):
  global inst
  relative = ''
  if statement.IsRelative:
    relative = 'RELATIVE'
  p = statement.Position.P
  e = calcRxRyRz( statement.Position )
  name = statement.Tool.Name
  inst.append('\'DEFINE %s TOOL %s %g,%g,%g,%g,%g,%g\n' %(relative,name,p.X,p.Y,p.Z,e.X,e.Y,e.Z))
#-------------------------------------------------------------------------------
def writeDelay(statement, ifclause=''):
  global inst
  inst.append('TIMER T=%g%s\n' % (statement.Delay,ifclause))
#-------------------------------------------------------------------------------
def writeGrasp(statement, ifclause=''):
  global inst
  inst.append('\'GRASP%s\n' % ifclause)
#-------------------------------------------------------------------------------
def writeHalt(statement, ifclause=''):
  global inst
  inst.append('PAUSE%s\n' % ifclause)
#-------------------------------------------------------------------------------
def writeHome(statement, ifclause=''):
  global inst
  inst.append('\'MOVE HOME%s\n' % ifclause)
#-------------------------------------------------------------------------------
def writeIf(statement, ifclause=''):
  global inst
  comprasions = ['==', '<>', '<=', '>=', '<', '>']
  cnd = statement.Condition.strip()
  sign_found = False
  for c in comprasions:
    if c in cnd:
      sign_found = True
      break
  if sign_found:
    element1, element2 = cnd.split(c)
    if c == '==':
      c = '='
  else:
    element1, element2 = cnd, 'True'
    c = '='
  ifclause = ' IF '+ element1 + c + element2 

  writeStatement = getAllStatementWriters()
  for statement in statement.ThenScope.Statements:
    writeStatement.get(statement.Type,unhandled)(statement, ifclause)
  for s in statement.ElseScope.Statements:
    print 'If %s: ElseScope is not supported.' % (cnd)
#-------------------------------------------------------------------------------
def writeLinMotion(statement, ifclause=''):
  global controller, motiontarget
  global usePL
  global posAxes, trackAxes
  global inst
  global posIndex

  statement.writeToTarget(motiontarget)
  controller.moveImmediate(motiontarget)
  
  writePosition(motiontarget)
  
  motion = 'MOVL'
  speed = '%0.1f' % statement.MaxSpeed
  jspeed = (100.0*motiontarget.JointSpeedFactor)
  
  pl = ''
  if usePL:
    if statement.Acceleration > EPSILON: # not continuous
      if statement.AccuracyValue > EPSILON:
        pl = ' PL=1'
      else:
        pl = ' PL=0'
  
  if trackAxes:
    bc = ' BC%05i' % posIndex
  else:
    bc = ''
    
  if posAxes:
    if useSyncMoves:
      if isBaseInStation(motiontarget):
        motion = 'SMOVL'
        ec = '  +MOVJ EC%05i' % (posIndex)
      else:
        motion = 'MOVL'
        ec = '  +MOVJ EC%05i VJ=%s' % (posIndex,jspeed)
    else:
      bc += ' EC%05i' % posIndex
      ec = ''
  else:
    ec = ''
   
  inst.append('%s C%05i%s V=%s%s%s%s\n' % (motion,posIndex,bc,speed,pl,ec,ifclause) )
#-------------------------------------------------------------------------------
def writePath(statement, ifclause=''):
  global posAxes, trackAxes
  global usePL
  global inst
  global motiontarget
  global posIndex

  inst.append('\'Move along path %s\n' % (statement.Name))

  motiontarget.JointTurnMode = VC_MOTIONTARGET_TURN_NEAREST
  motiontarget.TargetMode = VC_MOTIONTARGET_TM_NORMAL
  motiontarget.MotionType = VC_MOTIONTARGET_MT_LINEAR

  if statement.Base == None:
    motiontarget.BaseName = ''
  else:
    motiontarget.BaseName = statement.Base.Name
  if statement.Tool == None:
    motiontarget.ToolName = ''
  else:
    motiontarget.ToolName = statement.Tool.Name
  ej = statement.ExternalJointCount    

  for i in range( statement.getSchemaSize()):
    target = statement.getSchemaValue(i,'Position')
    motiontarget.Target = target
    
    writePosition(motiontarget)
    
    
    motion = 'MOVL'
    try:
      jv = motiontarget.JointValues
      for j in range(ej):
        jv[6+j] = stat.getSchemaValue(j,'E%d' % (j+1))
      motiontarget.JointValues = jv
      #
      pl = ''
      if usePL:
        if statement.getSchemaValue(i,'Acceleration') > EPSILON: # not continuous
          if statement.getSchemaValue(i,'AccuracyValue') > EPSILON:
            pl = ' PL=1'
          else:
            pl = ' PL=0'
      #
      speed = '%0.1f' % statement.getSchemaValue(i,'MaxSpeed')
      jspeed = (100.0*motiontarget.JointSpeedFactor)
    except:
      pl = ''
      #
      speed = '100.0'
      jspeed = '10.00'
      
    if trackAxes:
      bc = ' BC%05i' % posIndex
    else:
      bc = ''
    
    if posAxes:
      if useSyncMoves:
        if isBaseInStation(motiontarget):
          motion = 'SMOVL'
          ec = '  +MOVJ EC%05i' % (posIndex)
        else:
          motion = 'MOVL'
          ec = '  +MOVJ EC%05i VJ=%s' % (posIndex,jspeed)
      else:
        bc += ' EC%05i' % posIndex
        ec = ''
    else:
      ec = ''
    
    inst.append('%s C%05i%s V=%s%s%s%s\n' % (motion,posIndex,bc,speed,pl,ec,ifclause) )
  inst.append('\'End of path %s\n' % (statement.Name))
#-------------------------------------------------------------------------------
def writePosition( motiontgt ):
  global comp, controller
  global usePulses, useSyncMoves, convertToFrame
  global bases, tools
  global usedBases, usedTools
  global robotAxes, trackAxes, posAxes
  global Ccount, BCcount, ECcount 
  global cxxxxx, ecxxxxx, bcxxxxx
  global posIndex
  global initbase, inittool, prevtool, frame
  global pointsWereConverted

  #Incr index
  posIndex += 1
  
  #
  if not motiontgt.BaseName in usedBases:
    usedBases.append(motiontgt.BaseName)
  if not motiontgt.ToolName in usedTools:
    usedTools.append(motiontgt.ToolName)
  
  #Tool
  tool = getFrameIndex(motiontgt.ToolName,'TOOL',tools)
  if inittool < -1:
    inittool = tool
    prevtool = tool
  if tool != prevtool:
    cxxxxx.append('///TOOL %i\n' % tool)
    prevtool = tool
  
  #Robot position
  if usePulses:
    #PULSE
    jv = ''
    sep = '='
    for ii in range(robotAxes):
      jv += sep + '%i' % int(motiontgt.JointValues[ii]*ppr[ii]/360.0 + ppz[ii] )
      sep = ','
    cxxxxx.append('C%05i%s\n' % (posIndex,jv))
  else:
    #RECTAN
    base = getFrameIndex(motiontgt.BaseName, 'UFRAME', bases ) #-1=ROBOT,0:ROBOT,1=UFRAME1 etc.
    if initbase < -1:
      initbase = base
    #Conf
    writeRCONF( True )
    #Target
    if base != initbase or initbase <= 0:
      #Need to convert target into different base. *NULL* and UFRAME0 positions are always converted to correct robot world frame.
      if base != initbase:
        pointsWereConverted = True
      if initbase > 0:
        #initbase is UFRAME1-15 or custom base frame
        base = controller.Bases[initbase]
        if base.Node:
          bm = base.Node.WorldPositionMatrix * base.PositionMatrix
        else:
          #No base node, reference to robot world frame
          rm = comp.WorldPositionMatrix
          kin = comp.findBehaviour('Kinematics')
          if kin:
            offset_p = kin.getProperty('L01Z')
            if offset_p:
              rm.translateRel(0,0,offset_p.Value)
          bm = rm * base.PositionMatrix
      else:
        #initbase is *NULL* or UFRAME0, use robot world frame
        rm = comp.WorldPositionMatrix
        kin = comp.findBehaviour('Kinematics')
        if kin:
          offset_p = kin.getProperty('L01Z')
          if offset_p:
            rm.translateRel(0,0,offset_p.Value)
        bm = rm
      ibm = vcMatrix.new(bm)
      ibm.invert()
      tm = motiontarget.getSimWorldToRobotTool()
      m = ibm * tm
      p = m.P
      e = calcRxRyRz( m )
      p = zeroOut( p, 0.001 )
      e = zeroOut( e, 0.001 )
    else:
      #Use target in reference if base is not *NULL* or UFRAME0 and base is the same as initbase.
      p = motiontgt.Target.P
      e = calcRxRyRz( motiontgt.Target )
      p = zeroOut( p, 0.001 )
      e = zeroOut( e, 0.001 )
    cxxxxx.append('C%05i=%.3f,%.3f,%.3f,%.4f,%.4f,%.4f\n' % (posIndex,p.X,p.Y,p.Z,e.X,e.Y,e.Z))
  Ccount += 1

  #Track position
  if trackAxes:
    jv = ''
    sep = '='
    for ii in range(robotAxes, robotAxes+trackAxes):
      if usePulses:
        jv += sep + '%i' % int(motiontgt.JointValues[ii]*ppr[ii] + ppz[ii])
      else:
        jv += sep + '%.3f' % (motiontgt.JointValues[ii])
      sep = ','
    bcxxxxx.append('BC%05i%s\n' % (posIndex,jv))
    BCcount += 1

  #Positioner position
  if posAxes:
    jv = ''
    sep = '='
    for ii in range(robotAxes+trackAxes, robotAxes+trackAxes+posAxes):
      if usePulses:
        jv += sep + '%i' % int(motiontgt.JointValues[ii]*ppr[ii]/360.0 + ppz[ii] )
      else:
        jv += sep + '%.3f' % (motiontgt.JointValues[ii])
      sep = ','
    ecxxxxx.append('EC%05i%s\n' % (posIndex,jv))
    ECcount += 1
  
  return posIndex
#-------------------------------------------------------------------------------
def writePrint(statement, ifclause=''):
  global inst
  inst.append('PRINT \'%s\'%s\n' % (statement.Message,ifclause))
#-------------------------------------------------------------------------------      
def writeProgSync(statement, ifclause=''):
  global inst
  global tsync
  try:
    tsync = (int(statement.SyncMsg)-1)%32+1
  except:
    tsync = tsync%32 + 1
  inst.append('TSYNC %i\n' % tsync, ifclause )
#-------------------------------------------------------------------------------
def writePtpMotion(statement, ifclause=''):
  global controller, motiontarget
  global usePL
  global posAxes, trackAxes
  global inst
  global posIndex

  statement.writeToTarget(motiontarget)
  controller.moveImmediate(motiontarget)
  
  writePosition(motiontarget)
  
  motion = 'MOVJ'
  jspeed = (100.0*statement.JointSpeed)
  
  pl = ''
  if usePL:
    if statement.AccuracyValue > EPSILON:
      pl = ' PL=1'
    else:
      pl = ' PL=0'
  
  if trackAxes:
    bc = ' BC%05i' % posIndex
  else:
    bc = ''
    
  if posAxes:
    if useSyncMoves:
      ec = '  +MOVJ EC%05i VJ=%0.2f ' % (posIndex, jspeed )
    else:
      bc += ' EC%05i' % posIndex
      ec = ''
  else:
    ec = ''

  inst.append('%s C%05i%s VJ=%0.2f%s%s%s\n' % (motion,posIndex,bc,jspeed,pl,ec,ifclause))
#-------------------------------------------------------------------------------
def writeRelease(statement, ifclause=''):
  global inst
  inst.append('\'RELEASE%s\n' % ifclause)
#-------------------------------------------------------------------------------
def writeRemoteCall(statement, ifclause=''):
  global inst
  r = goodName(statement.RemoteRoutine.split(':')[2])
  remoteCount += 1
  remoteDict[r] = remoteCount
  inst.append('PSTART JOB:%s SUB%i%s\n' % (r,remoteCount,ifclause))
#-------------------------------------------------------------------------------
def writeRemoteWait(statement, ifclause=''):
  global inst
  r = goodName(statement.RemoteRoutine.split(':')[2])
  inst.append('PWAIT SUB%i%s\n' % (remoteDict.pop(r),ifclause) )
  remoteCount = 0
#-------------------------------------------------------------------------------
def writeSetBin(statement, ifclause=''):
  global inst
  if statement.OutputValue:
    value = 'ON'
  else:
    value = 'OFF'
  inst.append('DOUT OT#(%i) %s%s\n' %(statement.OutputPort,value,ifclause))
#-------------------------------------------------------------------------------
def writeSetProperty(statement, ifclause=''):
  value_expression = statement.ValueExpression.strip()
  inst.append('SET %s %s%s\n' % (statement.TargetProperty, value_expression,ifclause))
#-------------------------------------------------------------------------------  
def writeWaitBin(statement, ifclause=''):
  global inst
  if statement.InputValue:
    value = 'ON'
  else:
    value = 'OFF'
  inst.append('WAIT IN#(%i)=%s%s\n' %(statement.InputPort,value,ifclause))
#-------------------------------------------------------------------------------
def isBaseInStation(motiontgt):
  #Return true, if target's base moves with station. Else otherwise
  global controller
  global posCmp
  base = None
  for b in controller.Bases:
    if b.Name == motiontgt.BaseName:
      base = b
      break
  if not base:
    return False
  servos = posCmp.findBehavioursByType(VC_SERVOCONTROLLER)
  if len(servos) == 0:
    return False
  servo = servos[0]
  #
  node = base.Node
  while node:
    if node == servo.FlangeNode:
      return True
    node = node.Parent
  return False
#-------------------------------------------------------------------------------
def getJobName(name):
  if name == '':
    return 'MAIN'
  return goodName( name ) 
#-------------------------------------------------------------------------------
def goodName( inName ):
  badChars  = re.compile( '[^\w!%&\'()_-]' ) 
  outName = inName.strip()
  outName = outName.upper()
  outName = badChars.sub( '_', outName )
  if len(outName) > 32:
    print 'WARNING: Routine name <%s> exceeds 32 characters' % outName
  return outName
#-------------------------------------------------------------------------------
def getFrameIndex(str,frame,strs):
  if frame == str[:len(frame)]:
    return eval(str[len(frame):])
  idx = getStringIndex(str,strs)
  if frame == 'UFRAME' and idx == 0:
    idx = -1 #UFRAME0 does not exist in real robot, use ROBOT instead
  return idx
#-------------------------------------------------------------------------------
def getStringIndex(str,strs):
  i = 0
  for s in strs:
    if s == str:
      return i
    i += 1
  return -1
#-------------------------------------------------------------------------------
def zeroOut( p, tol ):
  if abs(p.X) < tol:
    p.X = 0.0
  if abs(p.Y) < tol:
    p.Y = 0.0
  if abs(p.Z) < tol:
    p.Z = 0.0

  return p
#-------------------------------------------------------------------------------
def calcRxRyRz( m ):
  #Set useEulerAngles True, if robot has eulers enabled
  useEulerAngles = False
  if useEulerAngles:
    A = atan2( m.A.Y, m.A.X )*RAD_TO_DEG
    B = asin( m.A.Z )*RAD_TO_DEG
    C = atan2( -m.O.Z, -m.N.Z )*RAD_TO_DEG
    return vcVector.new( A, B, C )
  else:
    return m.WPR
#-------------------------------------------------------------------------------
def writeRCONF( turns ):
  global prevrconf, prevt4, prevt6, cxxxxx, bcxxxxx

  rconf = motiontarget.RobotConfig

  jv = motiontarget.JointValues
  if turns and abs(jv[3]) > 180.0:
    t4 = 1
  else:
    t4 = 0
  if turns and abs(jv[5]) >= 180.0:
    t6 = 1
  else:
    t6 = 0

  if rconf == prevrconf and t4 == prevt4 and t6 == prevt6:
    return

  if rconf == 0:
    line = '///RCONF 1,0,0'
  elif rconf == 1:
    line = '///RCONF 0,0,0'
  elif rconf == 2:
    line = '///RCONF 1,1,0'
  elif rconf == 3:
    line = '///RCONF 0,1,0'
  elif rconf == 4:
    line = '///RCONF 1,0,1'
  elif rconf == 5:
    line = '///RCONF 0,0,1'
  elif rconf == 6:
    line = '///RCONF 1,1,1'
  elif rconf == 7:
    line = '///RCONF 0,1,1'

  line += ',%i,%i,0' %(t4,t6)

  if True:  #Controllers 'DX100', and 'DX200'
    line += ',0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0'
  line += '\n'

  cxxxxx.append(line)

  prevrconf = rconf
  prevt4 = t4
  prevt6 = t6
#-------------------------------------------------------------------------------
def unhandled(statement):
  print 'Unhandled statement : %s %s' % (statement.Name,statement.Type)
#------------------------------------------------------------------------------- 
def getAllStatementWriters():
  statementhandlers = {
  VC_STATEMENT_BREAK:unhandled,
  VC_STATEMENT_CALL:writeCall,
  VC_STATEMENT_COMMENT:writeComment,
  VC_STATEMENT_CONTINUE:unhandled,
  VC_STATEMENT_CUSTOM:unhandled,
  VC_STATEMENT_DEFINE_BASE:writeDefineBase,
  VC_STATEMENT_DEFINE_TOOL:writeDefineTool,
  VC_STATEMENT_DELAY:writeDelay,
  VC_STATEMENT_GRASP:writeGrasp,
  VC_STATEMENT_HALT:writeHalt,
  VC_STATEMENT_HOME:writeHome,
  VC_STATEMENT_IF:unhandled,
  VC_STATEMENT_LINMOTION:writeLinMotion,
  VC_STATEMENT_PATH:writePath,
  VC_STATEMENT_PRINT:writePrint,
  VC_STATEMENT_PROCESS:unhandled,
  VC_STATEMENT_PROG_SYNC:writeProgSync,
  VC_STATEMENT_PTPMOTION:writePtpMotion,
  VC_STATEMENT_RELEASE:writeRelease,
  VC_STATEMENT_REMOTECALL:writeRemoteCall,
  VC_STATEMENT_REMOTEWAIT:writeRemoteWait,
  VC_STATEMENT_RETURN:unhandled,
  VC_STATEMENT_SETBIN:writeSetBin,
  VC_STATEMENT_SETPROPERTY:writeSetProperty,
  VC_STATEMENT_WAITBIN:writeWaitBin,
  VC_STATEMENT_WHILE:unhandled,
  }
  return statementhandlers